from enum import Enum


class EnvironmentVariable(Enum):
    ENVIRONMENT: str = "ENVIRONMENT"
    AZURE_KEY_VAULT_URL: str = "AZURE_KEY_VAULT_URL"
    AZURE_CLIENT_ID: str = "AZURE_CLIENT_ID"
    AZURE_CLIENT_SECRET: str = "AZURE_CLIENT_SECRET"
    AZURE_TENANT_ID: str = "AZURE_TENANT_ID"


class EnvironmentSecret(Enum):
    APPLICATION_NAME: str = "APPLICATION-NAME"
    SENTRY_URL: str = "SENTRY-URL"
    MONGO_DB_CONNECTION_STRING: str = "MONGO-DB-CONNECTION-STRING"
    DATABASE_NAME: str = "DATABASE-NAME"
    DISCORD_BOT_TOKEN: str = "DISCORD-BOT-TOKEN"
    WISE_PRIMARY_ACCOUNT_API_KEY: str = "WISE-PRIMARY-ACCOUNT-API-KEY"
    WISE_SECONDARY_ACCOUNT_API_KEY: str = "WISE-SECONDARY-ACCOUNT-API-KEY"
    ENTRA_ID_CLIENT_ID: str = "ENTRA-ID-CLIENT-ID"
    ENTRA_ID_TENANT_ID: str = "ENTRA-ID-TENANT-ID"
    TWILIO_AUTH_TOKEN: str = "TWILIO-AUTH-TOKEN"
    TWILIO_ACCOUNT_SID: str = "TWILIO-ACCOUNT-SID"
    TWILIO_WHATSAPP_NUMBER: str = "TWILIO-WHATSAPP-NUMBER"
    TWILIO_SMS_NUMBER: str = "TWILIO-SMS-NUMBER"
