ACQUIRE_ACCESS_TOKEN__POLLING_SLEEP_SECONDS: int = 1
ACQUIRE_ACCESS_TOKEN__POLLING_TIMEOUT_SECONDS: int = 5 * 60
AUTHENTICATION_KEY_NAME: str = "KEY-AUTHENTICATION"
DEFAULT_ACCESS_TOKEN_VALIDITY_SECONDS: int = 3600
